.component{
    padding: 0px;
    background-color: white;
}

.sidebarGrammar{
    position: fixed; 
    z-index: 10;
    width: 25vw;
    height: 100vh;
    left: 0;
    top: 0;
    transition: 300ms cubic-bezier(.36,-0.01,0,.77);
}

.sidebarGrammar--isHidden{
    margin-left: -25vw;
}

.sidebarGrammar--isOpen{
    margin-left: 25vw;
}

.hidden{
    display: none;
}

#toggleSideBar{
    position: fixed;
    top: 0;
    left: 0;
    z-index: 11;
    border: 1px solid rgb(218, 220, 224);
    box-shadow: rgba(99, 99, 99, 0.2) 0px 2px 8px 0px;
    background-color: #fff;
    color: #fff;
    text-decoration: none;
    padding: 5px 10px;
    transition: 300ms cubic-bezier(.36,-0.01,0,.77);
}  