"""Rename profileImage to profile_image

Revision ID: 62a73b08f79f
Revises: 13c00188213d
Create Date: 2024-02-15 09:16:48.535205

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '62a73b08f79f'
down_revision = '13c00188213d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.add_column(sa.Column('profile_image', sa.String(length=200), nullable=True))
        batch_op.drop_column('profileImage')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.add_column(sa.Column('profileImage', sa.VARCHAR(length=200), nullable=True))
        batch_op.drop_column('profile_image')

    # ### end Alembic commands ###
