.logo {
  margin-right: 15px;
  margin-left: 15px;
  max-height: 100%;
  width: auto;
  filter: brightness(0) invert(1);
}

.menuBar {
  position: fixed;
  width: 100%;
  top: 0;
  display: flex;
  align-items: center;
  background-color: #1E1F23;
  height: 65px;
  padding: 10px;
  font-family: Rubik;
  z-index: 100;
  box-shadow: 0 6px 8px -4px rgba(0, 0, 0, 0.25);
}

.button {
  margin: 0 10px;
  padding: 6px 10px;
  cursor: pointer;
  font-weight: bold;
  background-color: transparent;
  color: #fbfcf6;
  border: none;
  /* border-bottom: 3px solid #fbfcf6; */
  border-radius: 4px;
}

.dashboardOn {
  color: rgb(251, 170, 105);
}

.dashboardOff {
  background: transparent;
}

.dropDownMenu {
  position: absolute;
  top: calc(100% + 20px);
  left: 10px;
  background-color: #1E1F23;
  box-shadow: 0 2px 5px rgba(0, 0, 0, 0.2);
  z-index: 100;
  width: 200px;
  border: 1px solid rgba(0, 0, 0, 0.2);
  border-radius: 0;
}

.dropDownRow {
  width: 100%;
  border: none;
  background-color: #1E1F23;
  padding: 5px;
  border-radius: 0;
  display: flex;
  flex-direction: row;
  cursor: pointer;
  align-items: center;
  justify-content: space-evenly;
}

.dropDownRow button {
  padding: 0;
}

.dropDownRow:hover {
  background-color: rgb(251, 170, 105);
}

.dropDownIcon {
  color: white;
}

.noStyleButton {
  background: none;
  border: none;
  color: white;
}

.workflowNameContainer {
  margin-top: 20px;
  text-align: center;
  z-index: 80;
  top: 60px;
  left: 50px;
  position: fixed;
}

.workflowNameStyle {
  font-size: 24px;
  cursor: pointer;
}

.input {
  font-size: 24px;
  text-align: center;
  border: 1px solid #ccc;
  border-radius: 4px;
  padding: 5px;
}

.dropdownWrapper {
  position: relative;
}

.rightSide { 
  display: flex;
  position: absolute;
  right: 90px;
  top: 12px;
  z-index: 200;
}

