"""Add provider and provider_uid to User

Revision ID: 13c00188213d
Revises: f5c32a160bb4
Create Date: 2024-02-15 09:13:27.568691

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '13c00188213d'
down_revision = 'f5c32a160bb4'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.add_column(sa.Column('provider', sa.String(length=50), nullable=True))
        batch_op.add_column(sa.Column('provider_uid', sa.String(length=200), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.drop_column('provider_uid')
        batch_op.drop_column('provider')

    # ### end Alembic commands ###
