.modal {
    position: fixed;
    display: flex;
    justify-content: center;
    flex-direction: column;
    align-items: center;
    top: calc(50% - 35%);
    left: 25%;
    width: 50%;
    height: 70%;
    background-color: white;
    box-shadow: rgba(0, 0, 0, 0.35) 0px 5px 15px;
    border-radius: 10px;
    z-index: 500;
}

.modalBackground {
    position: fixed;
    top: 0;
    left: 0;
    width: 100vw;
    height: 100vh;
    background-color: black;
    opacity: 50%;
    z-index: 400;
}

.closeX {
    position: absolute;
    right: 15px;
    top: 10px;
    cursor: pointer;
    font-weight: bold;
    font-size: 19px;
}

.title {
    font-weight: bold;
    font-size: 22px;
}

.graphDiv {
    width: 85%;
    height: 85%;
    position: relative;
    border: 1px solid black;
}